import msgpack
import pandas as pd
import uuid
import time
from datetime import timedelta, datetime
import threading
import random
import capnp
import data_capnp
from bw2python import ponames
from bw2python.bwtypes import PayloadObject
from bw2python.client import Client
from xbos.util import pretty_print_timedelta, TimeoutException
RAW  = 0
MEAN = 1 << 1
MIN = 1 <<  2
MAX = 1 << 3
COUNT = 1 << 4

DEFAULT_TIMEOUT=30
class BOSSWAVEMDALClient(object):
    """
    Simple Python client for the Metadata-Driven-Access-Layer (MDAL).
    MDAL ties together Brick models and timeseries data through the use of the `brickframe:uuid`
    relationship in the Brick model. The idea is this should replace tieing together HodDB
    and Pundat manually
    """

    def __init__(self, url, client=None, timeout=30):
        """
        Creates an MDAL client.

        Arguments:
        [url]: the BOSSWAVE uri where mdal is hosted
        [client]: if this is None, we use environment vars to configure a client
                  automatically; else, we use the client provided from bw2python
        """
        if client is None:
            client = Client()
            client.vk = client.setEntityFromEnviron()
            client.overrideAutoChainTo(True)
        if not isinstance(client, Client):
            raise TypeError("first argument must be bw2python.client.Client or None")
        self.c = client
        self.vk = client.vk
        self.url = url

        # check liveness
        responses = self.c.query("{0}/*/s.mdal/!meta/lastalive".format(url))
        for resp in responses:
            # get the metadata records from the response
            md_records = filter(lambda po: po.type_dotted == ponames.PODFSMetadata, resp.payload_objects)
            # get timestamp field from the first metadata record
            last_seen_timestamp = msgpack.unpackb(md_records[0].content).get('ts')
            # get how long ago that was
            now = time.time()*1e9 # nanoseconds
            # convert to microseconds and get the timedelta
            diff = timedelta(microseconds = (now - last_seen_timestamp)/1e3)
            print "Saw [{0}] MDAL {1}".format(self.url, pretty_print_timedelta(diff))
            if diff.total_seconds() > timeout:
                raise TimeoutException("MDAL at {0} is too old".format(self.url))

    def do_query(self, query, timeout=DEFAULT_TIMEOUT):
        """
        Query structure is as follows:

        query = {
            # We bind UUIDs found as the result of a Brick query to a variable name
            # that we can use later.
            # Each variable definition has the following:
            #  - name: how we will refer to this group of UUIDs
            #  - definition: a Brick query. The SELECT clause should return variables that end in '_uuid', which can be found as the
            #          object of a 'bf:uuid' relationship
            #  - units: what units we want to retrieve this stream as. Currently supports W/kW, Wh/kWh, F/C, Lux
            "Variables": [
                {"Name": "meter",
                 "Definition": "SELECT ?meter_uuid WHERE { ?meter rdf:type/rdfs:subClassOf* brick:Electric_Meter . ?meter bf:uuid ?meter_uuid . };",
                 "Units": "kW",
                },
                {"Name": "temp",
                 "Definition": "SELECT ?temp_uuid WHERE { ?temp rdf:type/rdfs:subClassOf* brick:Temperature_Sensor . ?temp bf:uuid ?temp_uuid . };",
                 "Units": "F",
                },
            ],

            # this is the composition of the data matrix we are returning. Below, all the uuids for the "meter" variable will be placed before
            # all of the uuids for the "temp" variable. We cannot guarantee order of uuids within those groups, but the ordering of the groups
            # will be preserved. Explicit UUIDs can also be used here
            "Composition": ["meter", "temp"],

            # If we are retrieving statistical data, then we need to say which statistical elements we want to download.
            # The options are RAW, MEAN, MIN, MAX and COUNT. To query multiple, you can OR them together (e.g. MEAN|MAX).
            # This maps 1-1 to the "Composition" field
            "Selectors": [MEAN, MEAN],

            # Themporal parameters for the query. Retrieves data in the range [T0, T1]. By convention, T0 < T1,
            # but MDAL will take care of it if this is reversed.
            # WindowSize is the size of the resample window in nanoseconds
            # if Aligned is true, then MDAL will snap all data to the begining of the window (e.g. if 5min window + Aligned=true,
            # then all timestamps will be on 00:05:00, 00:10:00, 00:15:00, etc)
            "Time": {
                "T0": "2017-08-01 00:00:00",
                "T1": "2017-08-08 00:00:00",
                "WindowSize": '2h',
                "Aligned": True,
            },
        }
        """
        nonce = str(random.randint(0, 2**32))
        query['Nonce'] = nonce
        ev = threading.Event()
        response = {}
        def _handleresult(msg):
            got_response = False
            for po in msg.payload_objects:
                if po.type_dotted == (2,0,10,4):
                    data = msgpack.unpackb(po.content)
                    if data['Nonce'] != query['Nonce']:
                        continue
                    print 'error' in data
                    if 'error' in data:
                        response['error'] = data['error']
                        got_response=True
                        continue
                    uuids = [str(uuid.UUID(bytes=x)) for x in data['Rows']]
                    data = data_capnp.StreamCollection.from_bytes_packed(data['Data'])
                    if hasattr(data, 'times'):
                        times = list(data.times)
                        df = pd.DataFrame(index=pd.to_datetime(times, unit='ns'))
                        for idx, s in enumerate(data.streams):
                            df[uuids[idx]] = s.values
                        response['df'] = df
                        got_response = True
                    else:
                        for idx, s in enumerate(data.streams):
                            if hasattr(s, 'times'):
                                s = pd.Series(s.values, pd.to_datetime(s.times, unit='ns'))
                            else:
                                s = pd.Series(s.values)
                            response[uuids[idx]] = s
                        got_response = True
            if got_response:
                ev.set()
        h = self.c.subscribe("{0}/s.mdal/_/i.mdal/signal/{1}".format(self.url, self.vk[:-1]), _handleresult)
        po = PayloadObject((2,0,10,3), None, msgpack.packb(query))
        self.c.publish("{0}/s.mdal/_/i.mdal/slot/query".format(self.url), payload_objects=(po,))
        ev.wait(timeout)
        self.c.unsubscribe(h)
        if 'error' in response:
            raise Exception(response['error'])
        return response

if __name__ == '__main__':
    query = {
        "Composition": ["meter", "temp"],
        "Selectors": [MEAN, MEAN],
        "Variables": [
            {"Name": "meter",
             "Definition": "SELECT ?meter_uuid WHERE { ?meter rdf:type/rdfs:subClassOf* brick:Electric_Meter . ?meter bf:uuid ?meter_uuid . };",
             "Units": "kW",
            },
            {"Name": "temp",
             "Definition": "SELECT ?temp_uuid WHERE { ?temp rdf:type/rdfs:subClassOf* brick:Temperature_Sensor . ?temp bf:uuid ?temp_uuid . };",
             "Units": "F",
            },
        ],
        "Time": {
            "T0": "2017-08-01 00:00:00",
            "T1": "2017-08-08 00:00:00",
            "WindowSize": int(1e9*60*5),
            "Aligned": True,
        },
        "Params": {
            "Statistical": False,
            "Window": True,
        },
    }

    c = MDALClient("scratch.ns")
    df = c.do_query(query)['df']
    print df.describe()



from xbos import get_client
from xbos.services.pundat import DataClient, timestamp, make_dataframe
from xbos.services.hod import HodClientHTTP

import brick

c = get_client()
dc = DataClient(c)

uri = "ciee/devices/venstar/s.venstar/OpenSpace/i.xbos.thermostat"

brick.get_thermostat_triples(dc, uri)



from bw2python.client import Client as BW2Client
import os

__all__ = ['devices','services','util']

def get_client(agent=None,entity=None):
    # set defaults
    if agent is None:
        agent = os.environ.get('BW2_AGENT','127.0.0.1:28589')
    if entity is None:
        entity = os.environ.get('BW2_DEFAULT_ENTITY',None)

    if agent is None:
        raise Exception("Need to provide an agent or set BW2_AGENT")
    if entity is None:
        raise Exception("Need to provide an entity or set BW2_DEFAULT_ENTITY")

    hostname, port = agent.split(':')
    c = BW2Client(hostname, int(port))
    c.setEntityFromFile(entity)
    c.overrideAutoChainTo(True)

    return c



"""
This library contains a set of utility functions for helping populate points related to smart devices

Given the BOSSWAVE URI for a device, it will look up in the archiver for the related timeseries points, using a regex
on the URI and knowledge of the standard xbos interface.

Each method generates a set of triples instantiating the device and adding the requisite edges for URI/UUID pointers
"""

from rdflib import Graph, Namespace, URIRef, Literal
RDF = Namespace('http://www.w3.org/1999/02/22-rdf-syntax-ns#')
RDFS = Namespace('http://www.w3.org/2000/01/rdf-schema#')
BRICK = Namespace('https://brickschema.org/schema/1.0.1/Brick#')
OWL = Namespace('http://www.w3.org/2002/07/owl#')
BF = Namespace('https://brickschema.org/schema/1.0.1/BrickFrame#')

i_xbos_thermostat_points = {
    'temperature': BRICK.Temperature_Sensor,
    'relative_humidity': BRICK.Relative_Humidity_Sensor,
    'heating_setpoint': BRICK.Supply_Air_Temperature_Heating_Setpoint,
    'cooling_setpoint': BRICK.Supply_Air_Temperature_Cooling_Setpoint,
    'override': BRICK.Override_Command,
    'fan': BRICK.Fan_Command,
    'state': BRICK.Thermostat_Status,
    'mode': BRICK.Thermostat_Mode_Command,
}


i_xbos_light_points = {
    'brightness': BRICK.Lighting_System_Luminance_Command,
    'state': BRICK.Lighting_State,
}

i_xbos_plug_points = {
    'state': BRICK.On_Off_Command,
    'power': BRICK.Electric_Meter,
}

class Generator:
    def __init__(self, bricknamespace, archiverclient):
        self.ns = bricknamespace
        self.archiver = archiverclient

    def add_xbos_thermostat(self, node, uri, controls=None):
        rest_of_uri = '/'.join(uri.split("/")[1:])
        namespace = uri.split("/")[0]
        md = self.archiver.query("select name, uuid where namespace = '{0}' and originaluri like '{1}'".format(namespace, rest_of_uri)).get('metadata')
        triples = []
        triples.append((node, RDF.type, BRICK.Thermostat))
        triples.append((node, BF.uri, Literal(uri)))

        triples.append((BRICK.Thermostat_Status, RDFS.subClassOf, BRICK.Status))
        triples.append((BRICK.Thermostat_Status, RDF.type, OWL.Class))
        triples.append((BRICK.Thermostat_Mode_Command, RDFS.subClassOf, BRICK.Command))
        triples.append((BRICK.Thermostat_Mode_Command, RDF.type, OWL.Class))

        if controls is not None:
            triples.append((node, BF.controls, controls))
        for doc in md:
            if doc['name'] in i_xbos_thermostat_points.keys():
                pointname = self.ns[node.split('#')[-1]+"_"+doc['name']]
                triples.append((pointname, RDF.type, i_xbos_thermostat_points[doc['name']]))
                triples.append((node, BF.hasPoint, pointname))
                triples.append((pointname, BF.uuid, Literal(doc['uuid'])))
        print triples
        return triples

    def add_xbos_light(self, node, uri, controls=None):
        rest_of_uri = '/'.join(uri.split("/")[1:])
        namespace = uri.split("/")[0]
        md = self.archiver.query("select name, uuid where namespace = '{0}' and originaluri like '{1}'".format(namespace, rest_of_uri)).get('metadata')
        print md
        triples = []
        triples.append((node, RDF.type, BRICK.Lighting_System))
        triples.append((node, BF.uri, Literal(uri)))
        if controls is not None:
            triples.append((node, BF.controls, controls))
        for doc in md:
            if doc['name'] in i_xbos_light_points.keys():
                pointname = self.ns[node.split('#')[-1]+"_"+doc['name']]
                triples.append((pointname, RDF.type, i_xbos_light_points[doc['name']]))
                triples.append((node, BF.hasPoint, pointname))
                triples.append((pointname, BF.uuid, Literal(doc['uuid'])))
        print triples
        return triples

    def add_xbos_plug(self, node, uri):
        rest_of_uri = '/'.join(uri.split("/")[1:])
        namespace = uri.split("/")[0]
        md = self.archiver.query("select name, uuid where namespace = '{0}' and originaluri like '{1}'".format(namespace, rest_of_uri)).get('metadata')
        triples = []
        triples.append((BRICK.PlugStrip, RDFS.subClassOf, BRICK.Equipment))
        triples.append((BRICK.PlugStrip, RDF.type, OWL.Class))
        triples.append((node, RDF.type, BRICK.PlugStrip))
        triples.append((node, BF.uri, Literal(uri)))
        for doc in md:
            if doc['name'] in i_xbos_plug_points.keys():
                pointname = self.ns[node.split('#')[-1]+"_"+doc['name']]
                triples.append((pointname, RDF.type, i_xbos_plug_points[doc['name']]))
                triples.append((node, BF.hasPoint, pointname))
                triples.append((pointname, BF.uuid, Literal(doc['uuid'])))
        print triples
        return triples



from setuptools import setup, find_packages

print find_packages()
setup(name='xbos',
      version='0.0.22',
      description='Aggregate wrapper for XBOS services and devices',
      url='https://github.com/SoftwareDefinedBuildings/XBOS',
      author='Gabe Fierro',
      author_email='gtfierro@cs.berkeley.edu',
      packages=find_packages(),
      data_files=[('xbos/services', ['xbos/services/data.capnp'])],
      install_requires=[
        'docker==2.5.1',
        'delorean==0.6.0',
        'msgpack-python==0.4.2',
        'bw2python>=0.6.1',
        'requests>=2.12.2',
        'python-dateutil>=2.4.2',
        'pandas>=0.20.1'
      ],
      zip_safe=False)




import msgpack
import shutil
import os
import docker
import time
import threading
import random
import requests
import json
from datetime import timedelta, datetime

from bw2python import ponames
from bw2python.bwtypes import PayloadObject
from bw2python.client import Client
from xbos.util import pretty_print_timedelta, TimeoutException

DEFAULT_TIMEOUT=30

class HodClientHTTP(object):
    def __init__(self, url="http://localhost:47808"):
        self.url = url
    def do_query(self, query, values_only=True):
        url = self.url+'/api/query'
        resp = requests.post(url, data=query)
        if not resp.ok:
            raise Exception("Query to {0} failed ({1})".format(self.url, resp.reason))
        d = resp.json()
        count = d['Count']
        elapsed = d['Elapsed']
        rows = d['Rows']
        if values_only:
            rows = [{k: v['Value'] for k,v in r.items()} for r in rows]
        return rows

class HodClient(object):
    def __init__(self, url, client=None, timeout=20):
        if client is None:
            client = Client()
            client.vk = client.setEntityFromEnviron()
            client.overrideAutoChainTo(True)
        if not isinstance(client, Client):
            raise TypeError("first argument must be bw2python.client.Client or None")
        self.c = client
        self.vk = client.vk
        self.url = url

        responses = self.c.query("{0}/*/s.hod/!meta/lastalive".format(url))
        for resp in responses:
            # get the metadata records from the response
            md_records = filter(lambda po: po.type_dotted == ponames.PODFSMetadata, resp.payload_objects)
            # get timestamp field from the first metadata record
            last_seen_timestamp = msgpack.unpackb(md_records[0].content).get('ts')
            # get how long ago that was
            now = time.time()*1e9 # nanoseconds
            # convert to microseconds and get the timedelta
            diff = timedelta(microseconds = (now - last_seen_timestamp)/1e3)
            print "Saw [{0}] HodDB {1}".format(self.url, pretty_print_timedelta(diff))
            if diff.total_seconds() > timeout:
                raise TimeoutException("HodDB at {0} is too old".format(self.url))

    def do_query(self, query, timeout=DEFAULT_TIMEOUT, values_only=True):
        nonce = str(random.randint(0, 2**32))
        ev = threading.Event()
        response = {}
        def _handleresult(msg):
            # decode, throw away if not correct nonce
            got_response = False
            # if got response
            for po in msg.payload_objects:
                if po.type_dotted == (2,0,10,2):
                    data = msgpack.unpackb(po.content)

                    if data["Nonce"] != nonce:
                        continue
                    data.pop("Nonce")

                    if data["Error"] is not None and len(data["Error"]) > 0:
                        raise Exception(data["Error"])
                    else:
                        data.pop("Error")

                    for k,v in data.items():
                        response[k] = v
                    if values_only and response["Count"] > 0:
                        response["Rows"] = [{k: v["Value"] for k,v in r.items()} for r in response["Rows"]]
                    got_response = True
            if got_response:
                ev.set()
        h = self.c.subscribe("{0}/s.hod/_/i.hod/signal/result".format(self.url), _handleresult)
        q_struct = msgpack.packb({"Query": query, "Nonce": nonce})
        po = PayloadObject((2,0,10,1), None, q_struct)
        self.c.publish("{0}/s.hod/_/i.hod/slot/query".format(self.url), payload_objects=(po,))
        ev.wait(timeout)
        self.c.unsubscribe(h)
        if len(response) == 0: # no results
            raise TimeoutException("Query of {0} timed out".format(query))

        return response

class LocalHodDB(HodClientHTTP):
    def __init__(self, buildings):
        print buildings

        # make config directory
        shutil.rmtree('/tmp/_hoddb', ignore_errors=True)
        os.makedirs('/tmp/_hoddb')
        _build = ""
        for dbname, filename in buildings.items():
            newdest = os.path.join('/tmp/_hoddb', os.path.basename(filename))
            mountname = os.path.join('/etc/hod', os.path.basename(filename))
            shutil.copy(filename, newdest)
            _build += "    {0}: {1}\n".format(dbname, mountname)
        with open('/tmp/_hoddb/hodconfig.yaml','wb') as f:
            f.write(_CONFIG.format(_build))

        self.client = docker.from_env()
        # pull latest hod
        self.client.images.pull('gtfierro/hod:0.5.6')
        try:
            c = self.client.containers.get('localhod')
            c.kill()
            c.remove()
        except docker.errors.NotFound:
            pass

        self.container = self.client.containers.run('gtfierro/hod:0.5.6',
                                                volumes={'/tmp/_hoddb': {'bind': '/etc/hod',' mode': 'rw'}},
                                                name='localhod',
                                                detach=True)
        while True:
            self.container.reload()
            if '0.0.0.0:47808' in self.container.logs(tail='all'):
                break
        print 'ip addr', self.container.attrs['NetworkSettings']['Networks']['bridge']['IPAddress']
        self.ip = self.container.attrs['NetworkSettings']['Networks']['bridge']['IPAddress']
        #print dir(container)
        #container.logs(follow=True)
        HodClientHTTP.__init__(self, url="http://{0}:47808".format(self.ip))

_CONFIG = """
### HodDB Configuration
# commented out lines are default configuration options
Buildings:
    {0}

# the path to the TTL file containing Brick relationships
BrickFrameTTL: "/BrickFrame.ttl"
# the path to the TTL file containing Brick classes
BrickClassTTL: "/Brick.ttl"
DisableQueryCache: true

####
# Interface Enabling
####
# Enable HTTP server
EnableHTTP: true
# Enable BOSSWAVE server
EnableBOSSWAVE: false

####
# configuration for verbosity during operation
####
# Show the full latency of the query (and its larger components)
ShowQueryLatencies: true
# Set log level. In order of increasing verbosity:
# Debug, Info, Notice, Warning, Error, Critical
LogLevel: Debug


####
# HTTP Server Configuration
####
# port to run the server on
ServerPort: 47808
# Whether or not to serve on localhost. If false, serves on a public interface
ListenAddress: 0.0.0.0
# Path to the server directory of hod, which contains the necessary HTML files
StaticPath: /server

# If specified, serve the frontend over HTTPS using golang.org/x/crypto/acme/autocert
# If left blank (default), just serve over HTTP
#TLSHost: ""

####
# BOSSWAVE Server Configuration
####
# Base URI
HodURI: scratch.ns/hod
"""
